import TransitionSystem._
import Expr._

object Adder {
  def accId(i: Int): StateIdentifier = StateIdentifier(s"acc$i")
  def counterId(i: Int): StateIdentifier = StateIdentifier(s"counter$i")
  val overflow: StateIdentifier = StateIdentifier("overflow")
  def carryId(i: Int): AuxiliaryIdentifier = AuxiliaryIdentifier(s"carry$i")
  def inputId(i: Int): InputIdentifier = InputIdentifier(s"input$i")

  val invariant: Expr = Not(Var(overflow))

  // The adder is a transition system.
  // * It maintains an accumulator with `n` bits
  // * It takes as input `m` bits (with m <= n) which are to be added to the accumulator
  // * It maintains a counter with `l` bits, counting the number of inputs that have been received
  // * When the counter is full (all booleans are true), the adder just loops in the same state and stops accumulating inputs
  // * It maintains a boolean variable `overflow` which is set to true when the accumulator overflows
  def adder(n: Int, m: Int, l: Int): TransitionSystem = {
    require(m <= n)
    val accIds: Seq[StateIdentifier] = (0 to n-1).map(accId)
    val counterIds: Seq[StateIdentifier] = (0 to l-1).map(counterId)
    val carryIds: Seq[AuxiliaryIdentifier] = (0 to n).map(carryId)
    val inputIds: Seq[InputIdentifier] = (0 to m-1).map(inputId)
    val stateIds: Seq[StateIdentifier] = accIds ++ counterIds :+ overflow

    val fullCounter = and(counterIds.map(Var(_)))
    val computeCarry: Map[Identifier, Expr] =
      (1 to n).map { i =>
        val inputBit = if (i <= m) Var(inputId(i-1)) else BooleanLiteral(false)
        carryId(i) ->
          And(Not(fullCounter),
            or(
              And(Var(accId(i-1)), inputBit),
              And(Var(accId(i-1)), Var(carryId(i-1))),
              And(inputBit, Var(carryId(i-1))),
            )
          )
      }.toMap[Identifier, Expr].updated(
        carryId(0),
        BooleanLiteral(false)
      )

    val computeAcc: Map[Identifier, Expr] =
      (0 to n-1).map { i =>
        accId(i) ->
          Or(
            And(
              fullCounter,
              Var(accId(i))
            ),
            And(
              Not(fullCounter),
              xor(
                Var(accId(i)),
                if (i < m) Var(inputId(i)) else BooleanLiteral(false),
                Var(carryId(i))
              )
            )
          )
      }.toMap

    val computeCount: Map[Identifier, Expr] =
      (0 to l-1).map { i =>
        counterId(i) ->
          Or(
            fullCounter,
            And(
              Not(fullCounter),
              Equal(
                or(counterIds.take(i).map(id => Not(Var(id)))),
                Var(counterId(i))
              )
            )
          )
      }.toMap

    val computeOverflow: Map[Identifier, Expr] = Map(
      overflow -> Or(Var(overflow), Var(carryId(n)))
    )

    TransitionSystem(
      stateIds,
      carryIds,
      inputIds,
      computeAcc ++ computeCarry ++ computeCount ++ computeOverflow
    )
  }

  def mapToBigInt[T](n: Int, m: Map[T, Boolean], f: Int => T): BigInt = {
    BigInt((n-1 to 0 by -1).map(i => if (m(f(i))) '1' else '0').mkString, 2)
  }

  def overflowToBigInt(state: State): BigInt = mapToBigInt(1, state, _ => overflow)
  def stateToBigInt(n: Int, state: State): BigInt = mapToBigInt(n, state, accId)
  def inputToBigInt(n: Int, input: Input): BigInt = mapToBigInt(n, input, inputId)

  def bigIntToMap[T](n: Int, a: BigInt, f: Int => T): Map[T, Boolean] = {
    a.toString(2).reverse.padTo(n, '0').zipWithIndex.map { case (c,i) =>
      (f(i), c == '1')
    }.toMap
  }

  def interactiveTest(n: Int, m: Int, l: Int) = {
    val adder: TransitionSystem = Adder.adder(n, m, l)
    adder.checkWF

    def loop(state: State): Unit = {
      println(stateToString(n, l, state))
      println()
      val i = scala.io.StdIn.readLine("Enter 'STOP' or an input: ").trim
      println()
      if (i != "STOP") {
        val input = bigIntToMap(m, BigInt(i), inputId)
        loop(adder.next(state, input))
      }
    }
    loop(adder.stateIds.map(id => id -> false).toMap)
  }

  def stateToBigInts(n: Int, l: Int, s: State): (BigInt, BigInt, BigInt) = {
    val overflowBit = overflowToBigInt(s.filter(_._1 == overflow))
    val accumulator = stateToBigInt(n, s.filter(_._1.id.startsWith("acc")))
    val counter = mapToBigInt(l, s.filter(_._1.id.startsWith("counter")), counterId)

    (accumulator, counter, overflowBit)
  }

  def stateToString(n: Int, l: Int, s: State): String = {
    val (accumulator, counter, overflowBit) = stateToBigInts(n, l, s)

    s"Accumulator: $accumulator (${accumulator.toString(2).reverse.padTo(n, '0')})\n" +
    s"Counter:     $counter (${counter.toString(2).reverse.padTo(n, '0')})\n" +
    s"Overflow:    $overflowBit"
  }

  def pathToString(n: Int, m: Int, l: Int, path: (Seq[State], Seq[Input])): String = {
    val (sSeq, aSeq) = path
    sSeq.zip(aSeq).map { case (s,a) =>
      "State:" + "\n" +
      stateToString(n, l, s) + "\n\n" +
      "Input: " + inputToBigInt(m, a) + "\n"
    }.mkString("\n\n") + "\n" +
    "Final state:\n" +
    stateToString(n, l, sSeq.last) + "\n\n" +
    " => Overflow"
  }

  def checkNoOverflow(n: Int, m: Int, l: Int, debug: Boolean = true): Boolean = {
    val adder = Adder.adder(n, m, l)
    val initial: Expr = stateToExpr(adder.stateIds.map(id => id -> false).toMap)
    val maxSteps: Int = BigInt(2).pow(l).toInt
    adder.checkErrorReachabilityAtMost(initial, Not(Adder.invariant), maxSteps) match {
      case None =>
        if (debug)
          println(" => No overflow")
        true

      case Some(path) =>
        if (debug) {
          println(pathToString(n, m, l, path))
          println(" => Overflow")
        }
        false
    }
  }

  def lt(a: Expr, b: Expr): Expr = And(Not(a), b)

  // `as <= bs`
  // When a bit `i` in `as` is strictly greater than in `bs`, it must be that
  // there exists a more significant bit `j` in `bs` which is greater
  def leq(as: Seq[Expr], bs: Seq[Expr]): Expr = {
    require(as.size == bs.size)

    and((0 to as.size - 1).map(i =>
      Implies(
        lt(bs(i), as(i)),
        or((i + 1 to as.size - 1).map(j =>
          lt(as(j), bs(j))
        ))
      )
    ))
  }

  def inductiveInvariant(n: Int, m: Int, l: Int): Expr = {
    // acc <= 2^(n-l)*counter
    val acc = (0 to n-1).map(i => Var(accId(i)))
    val paddedCounter: Seq[Expr] =
      (1 to n-l).map(_ => BooleanLiteral(false)) ++
      (0 to l-1).map(i => Var(counterId(i)))

    And(invariant, leq(acc, paddedCounter))
  }
}
