import Adder._
import TransitionSystem._
import Utils._
import Expr._

object GradingLabs02 {

  /** Tests for Problem 1: Adder **/

  def noOverflowParameters(n: Int, m: Int, l: Int): Boolean = {
    (n == m && l == 1) ||
    (n == l && m == 1) ||
    n >= m + l
  }

  def adderTests(numInputs: Int, n: Int, m: Int, l: Int, ignoreFullCounter: Boolean): Unit = {
    require(m < 32)
    val adder = StudentToGrade.adder(n, m, l)
    var overflow: Boolean = false
    var state: State = adder.stateIds.map(id => id -> false).toMap
    var acc: BigInt = 0
    var counter: BigInt = 0
    var i = 0

    val maxInput = (BigInt(2).pow(m) - 1).toInt
    val maxAcc = BigInt(2).pow(n) - 1
    val maxCounter = BigInt(2).pow(l) - 1

    var history: List[BigInt] = List()

    var adderHadOverflow = false

    while (!overflow && i < numInputs) {
      i += 1
      val a: Int = util.Random.nextInt(maxInput + 1)
      val input: Input = bigIntToMap(m, a, inputId)
      history = a :: history
      if (counter < maxCounter) {
        acc += a
        counter += 1
        if (acc > maxAcc) {
          overflow = true
        }
      }
      state = adder.next(state, input)
      val (adderAcc, adderCounter, adderOverflow) = stateToBigInts(n, l, state)
      adderHadOverflow = adderHadOverflow || (adderOverflow == 1)

      if ((!ignoreFullCounter || counter != maxCounter) && (adderHadOverflow != overflow || (!overflow &&
        (acc != adderAcc || counter != adderCounter)
      ))) {
        println("History of inputs: " + history.reverse)
        println(s"(Adder state) Acc: $adderAcc, Counter: $adderCounter, Had Overflow: $adderHadOverflow")
        println(s"(Real state) Acc: $acc, Counter: $counter, Overflow: $overflow")
        throw new Exception(s"Problem during adder tests $n $m $m")
      }
    }
  }

  def adderTests(ignoreFullCounter: Boolean): Unit = {
    for (n <- 2 to 16) {
      println(s"Adder tests (ignore full counter: $ignoreFullCounter): accumulator bits: $n")
      for (m <- 2 to n) {
        for (l <- 2 to 16) {
          adderTests(200, n, m, l, ignoreFullCounter)
        }
      }
    }
  }

  def adderModelChecking(n: Int, m: Int, l: Int): Unit = {
    println(s"Model checking $n $m $l")
    val adder = StudentToGrade.adder(n, m, l)
    val initial: Expr = stateToExpr(adder.stateIds.map(id => id -> false).toMap)
    val maxSteps: Int = BigInt(2).pow(l).toInt
    val res = adder.checkErrorReachabilityAtMost(initial, Not(Adder.invariant), maxSteps)
    (res, noOverflowParameters(n, m, l)) match {
      case (None, false) =>
        throw new Exception(s"Model checking said no overflow on $n $m $l, while there should be one")
      case (Some(path), true) => ()
        throw new Exception(s"On parameters $n $m $l, model checking said overflow with path:\n${pathToString(n, m, l, path)}")
      case _ => ()
    }
  }

  def adderModelChecking(): Unit = {
    for (n <- 1 to 5) {
      println(s"Adder model checking: accumulator bits: $n")
      for (m <- 1 to n) {
        for (l <- 1 to 5) {
          adderModelChecking(n, m, l)
        }
      }
    }
  }

  def adderSimpleTest(): Unit = {
    val adder = StudentToGrade.adder(2, 2, 1)
    var state: State = adder.stateIds.map(id => id -> false).toMap
    val s0 = adder.next(state, bigIntToMap(2, 3, inputId))
    val s1 = adder.next(s0, bigIntToMap(2, 3, inputId))
    val overflowBit = s1(overflow)
    assert(!overflowBit, "Simple test failed")
    println("Simple test ok")
  }

  def evaluateAdder(): Unit = {
    adderTests(true)
    adderSimpleTest()
    adderModelChecking()
    adderTests(false)
  }


  /** Tests for Problem 2: Bounded Model Checker **/

  def checkPathValidity(ts: TransitionSystem, from: Expr, to: Expr, sSeq: Seq[State], aSeq: Seq[Input]): Boolean = {
    require (sSeq.length == aSeq.length + 1)

    val first = sSeq.head
    val last = sSeq.last
    from.interpret(first.map { case (k, v) => (k: Identifier ,v) }) &&
    to.interpret(last.map { case (k, v) => (k: Identifier ,v) }) &&
    ((0 to aSeq.length - 1).forall { i =>
      sSeq(i + 1) == ts.next(sSeq(i), aSeq(i))
    })
  }

  def reachabilityChallenge(
    stateIds: Seq[StateIdentifier],
    auxiliaryIds: Seq[AuxiliaryIdentifier],
    inputIds: Seq[InputIdentifier],
  ): Unit = {
    val ts = TransitionSystem.random(stateIds, auxiliaryIds, inputIds)
    val from = stateToExpr(randomMap(stateIds))
    val error = Expr.random(stateIds.toArray)

    for (i <- 1 to 5) {
      val ourResult = ts.checkErrorReachability(from, error, i)
      val theirResult = ts.studentCheckErrorReachability(from, error, i)
      ourResult match {
        case Some((sSeq, aSeq)) =>
          checkPathValidity(ts, from, error, sSeq, aSeq)
        case _ => ()
      }
      theirResult match {
        case Some((sSeq, aSeq)) =>
          checkPathValidity(ts, from, error, sSeq, aSeq)
        case _ => ()
      }
      (ourResult, theirResult) match {
        case (None, Some((sSeq, aSeq))) =>
          println(ts)
          println("Error: ", error)
          println("From: ", from)
          ts.dump("ts-path", Some((sSeq.zip(aSeq), sSeq.last)))
          throw new Exception("""Student's code said "reachable" while ours said "not reachable"""")
        case (Some((sSeq, aSeq)), None) =>
          println(ts)
          println("From: ", from)
          println("Error: ", error)
          ts.dump("ts-path", Some((sSeq.zip(aSeq), sSeq.last)))
          throw new Exception("""Our code said "reachable" while student's code said "not reachable"""")
        case _ =>
          ()
      }
    }

  }

  def modelCheckerTests() = {
    val stateIds = (1 to 9).map(i => StateIdentifier(s"x$i"))
    val auxiliaryIds = (1 to 9).map(i => AuxiliaryIdentifier(s"aux$i"))
    val inputIds = (1 to 9).map(i => InputIdentifier(s"input$i"))
    for (i <- 1 to 100) {
      reachabilityChallenge(stateIds, auxiliaryIds, inputIds)
      if (i % 10 == 0)
        println(s"Did $i successful tests.")
    }
  }

  // To evaluate students bounded model checker, please copy their code in
  // function `studentCheckErrorReachability` of file `TransitionSystem.scala`
  def evaluateModelChecker() = {
    modelCheckerTests()
  }


  /** Tests for Problem 3: Adder inductive invariant **/
  def inductiveInvariantTest(n: Int, m: Int, l: Int, target: Expr): Boolean = {
    val adder = Adder.adder(n, m, l)
    val studentsInductiveInvariant: Expr = StudentToGrade.inductiveInvariant(n, m, l)
    // println(studentsInductiveInvariant)
    val initial: Expr = stateToExpr(adder.stateIds.map(id => id -> false).toMap)
    adder.checkInductiveInvariant(studentsInductiveInvariant, initial, target)
  }

  def evaluateInductiveInvariant() = {
    val edge1 = inductiveInvariantTest(4, 1, 4, Adder.invariant)
    val edge2 = inductiveInvariantTest(4, 1, 4, BooleanLiteral(true))
    println("Edge case 1: " + edge1)
    println("Edge case 2: " + edge2)

    val large1 = inductiveInvariantTest(64, 32, 32, Adder.invariant)
    val large2 = inductiveInvariantTest(64, 32, 32, BooleanLiteral(true))
    println("Large case: " + large1)
    println("Large case: " + large2)
    // // inductiveInvariantTest(32, 16, 16)
    // // inductiveInvariantTest(8, 4, 4)
    var total = 0
    var succeeded = 0
    for (n <- 1 to 8) {
      println(s"Adder inductive invariant: accumulator bits: $n")
      for (m <- 1 to n) {
        for (l <- 1 to 8) {
          if (noOverflowParameters(n, m, l) && n >= m + l) {
            total += 1
            if (inductiveInvariantTest(n, m, l, Adder.invariant))
              succeeded += 1
          }
        }
      }
    }
    println("Success on non-edge case parameters (with target): " + (succeeded * 100.0 / total) + "%")

    var total2 = 0
    var succeeded2 = 0
    for (n <- 1 to 8) {
      println(s"Adder inductive invariant: accumulator bits: $n")
      for (m <- 1 to n) {
        for (l <- 1 to 8) {
          if (noOverflowParameters(n, m, l) && n >= m + l) {
            total2 += 1
            if (inductiveInvariantTest(n, m, l, BooleanLiteral(true)))
              succeeded2 += 1
          }
        }
      }
    }
    println("Success on non-edge case parameters (without target): " + (succeeded2 * 100.0 / total2) + "%")

    // println("total", total)
    // println("total2", total2)
  }
}
