import Utils._
import Tseytin._
import scala.annotation.tailrec

object SatSolver {

  import Solver._

  object Solver {
    def apply(cnf: CNF): Solver = {
      new {
        val numVars: Int = cnf.maxId
        val clauses: Array[Clause] = cnf.clauses
        val numClauses: Int = clauses.size
      } with Solver
    }

    def checkSat(e: Expr): Option[Map[Identifier, Boolean]] = {
      val (cnf, idToVar) = CNF.makeCNF(tseytin(e))
      checkSat(cnf).map(array =>
        idToVar.mapValues(x => int2Bool(array(x.id)))
      )
    }

    def checkSat(cnf: CNF): Option[Array[Int]] = Bench.time("checkSat") {
      val solver = Solver(cnf)
      solver.init()

      while (true) {
        // We propagate whatever can be propagated
        solver.propagate()

        if (!solver.hasConflict) {
          if (solver.allAssigned) {
            // If we don't have a conflict and all variables are assigned,
            // we have found a satisfying assignment.
            return Some(solver.assignments)
          } else {
            // If some variables remain unassigned, we need to make a new
            // decision.
            assert(solver.nextDecision.nonEmpty, "We should have a nextDecision here.")
            solver.decide(solver.nextDecision.get)
          }
        }
        else {
          // If we have a conflict, we need to backtrack
          solver.backtrack()
          // If `backtrack` did not set `nextDecision`, it means we have
          // explored all assignments and the formula is unsat.
          if (solver.nextDecision.isEmpty)
            return None
        }
      }

      sys.error("Cannot exit while(true) loop.")
    }
  }

  sealed abstract class Origin
  case object Decision extends Origin
  case object Propagation extends Origin

  trait Solver {

    /** These constants do not change during solving **/
    val numVars: Int
    val clauses: Array[Clause]
    val numClauses: Int

    require(numVars > 0)


    /** These are mutable and are modified during solving **/

    // The stack of assignments that we have made so far (can be backtracked
    // by the `backtrack` function).
    // For each literal, we remember whether it was set by a decision or by a
    // propagation.
    var trail: List[(Lit, Origin)] = List()

    // `assignments(i) =  0` means variable i is not assigned
    // `assignments(i) =  1` means variable i is assigned to `true`
    // `assignments(i) = -1` means variable i is assigned to `false`
    val assignments: Array[Int] = Array.ofDim(numVars + 1)

    // For each clause, we maintain the two literals that are watched
    var watchedLiterals: Array[Set[Lit]] = Array.ofDim(numClauses)

    // `watchingClauses(x)` returns the set of clauses indices that watch `x`.
    val watchingClauses: Array[Set[Int]] = Array.fill(numVars + 1)(Set())

    // This variable is set to `true` when a conflict is found
    var conflict: Boolean = false

    // This variable sets the next decision to make
    var nextDecision: Option[Lit] = Some(VarId(1).toLiteral(true))

    // During initialization,
    // * we make sure there are no empty clauses, and
    // * we initialize `watchedLiterals` and `watchingClauses`.
    def init(): Unit = {
      // No empty clause
      for (i <- 0 to clauses.length - 1) {
        assert(!clauses(i).isEmpty, "Empty clause means unsat")
      }

      // initialization of `watchedLiterals` and `watchingClauses`
      for (i <- 0 to clauses.length - 1) {
        watchedLiterals(i) = clauses(i).literals.toSet.take(2)
        for (lit <- watchedLiterals(i)) {
          watchingClauses(lit.getVar.id) = watchingClauses(lit.getVar.id) + i
        }
      }
    }

    def assignment(x: VarId): Int = assignments(x.id)

    def assume(lit: Lit, origin: Origin): Unit = {
      require (unassigned(lit), "We should only assign non-assigned literals")
      trail = (lit, origin) :: trail
      assignments(lit.getVar.id) = lit.positiveValue
    }

    def unassign(lit: Lit): Unit = {
      assignments(lit.getVar.id) = 0
    }

    def hasConflict: Boolean = conflict

    def unassigned(lit: Lit): Boolean = {
      assignments(lit.getVar.id) == 0
    }

    def isWellAssigned(lit: Lit): Boolean = {
      assignments(lit.getVar.id) == lit.positiveValue
    }

    def isBadlyAssigned(lit: Lit): Boolean = {
      assignments(lit.getVar.id) == -lit.positiveValue
    }

    /** Stop watching literal `lit` in clause `i`,
      * by updating `watchingClauses` and `watchedLiterals`.
      */
    def stopWatching(i: Int, lit: Lit): Unit = {
      watchingClauses(lit.getVar.id) -= i
      watchedLiterals(i) -= lit
    }

    /** Start watching literal `lit` in clause `i`,
      * by updating `watchingClauses` and `watchedLiterals`.
      **/
    def startWatching(i: Int, lit: Lit): Unit = {
      watchingClauses(lit.getVar.id) += i
      watchedLiterals(i) += lit
    }

    def propagate(): Unit = {
      propagate((0 to clauses.length - 1).flatMap(propagateClause).toSet)
    }

    /** This function explores clause `i` and checks if there is something to propagate.
      *
      * It considers three main cases:
      *
      *  - If a literal of `clauses(i)` is already `true`, this function returns
      *   `None` and does nothing.
      *
      * -  If there is a single literal `l1` watched by `i` which is unassigned: we
      *    consider two cases:
      *
      *   + Either we find another literal `l2` in `clauses(i)` which is unassigned
      *     and not watched yet by clause `i`. In which case, clause `i` starts
      *     watching it (`startWatching`), and clause `i` stops watching all
      *     literals different from `l1` and `l2`. The function returns None
      *     (nothing to propagate yet).
      *
      *   + Otherwise, we return `Some(l1)`. This means that `l1` has to be true
      *     and that we want to propagate it.
      *
      * - Otherwise, we do nothing and return `None`.
      */
    def propagateClause(i: Int): Option[Lit] = Bench.time("propagateClause") {
      val unassignedWatched = watchedLiterals(i).filter(unassigned)

      if (clauses(i).literals.exists(isWellAssigned)) {
        // this clause is already solved, we do nothing
        None
      }
      else if (unassignedWatched.size == 1) {
        // there is only one unassigned literal,
        // so we look for another unassigned literal to watch
        clauses(i).literals.find(l =>
          unassigned(l) && !watchedLiterals(i).contains(l)
        ) match {
          case None =>
            // There are no more literals to watch, we can propagate the remaining literal
            Some(unassignedWatched.head)
          case Some(newLit) =>
            // We found a new literal to watch
            for (l <- watchedLiterals(i) if unassignedWatched.head != l)
              stopWatching(i, l)

            startWatching(i, newLit)
            assert(watchedLiterals(i).size == 2, "The size should be 2 here")
            // Nothing to propagate here, because this clause still has at
            // least two unassigned literals
            None
        }
      } else {
        None
      }
    }

    /** This function takes as argument a set of literals to propagate.
      *
      * It maintains a set of literals to propagate, which is initially `toPropagate`.
      *
      * Iteratively, it pops a literal from the set of remaining literals to
      * propagate and does the following:
      *
      * - If the literal has already been set to the wrong value, the function
      *   sets the property `conflict` to `true` and continues to propagate the
      *   remaining literals.
      *
      * - If the literal is unassigned, we call `assume` to set its value,
      *   and we explore the clauses watching the corresponding variable with
      *   `propagateClause` to see if there are more things to propagate
      *   (these need to be added to the set of remaining literals to propagate)
      *
      * - If the literal is already assigned to the right value, we continue
      *   with the remaining literals
      */
    def propagate(toPropagate: Set[Lit]): Unit = Bench.time("propagate") {
      var remaining = toPropagate

      while (!remaining.isEmpty) {
        val lit = remaining.head
        val x = lit.getVar
        remaining = remaining.tail

        if (assignment(x) == -lit.positiveValue) {
          conflict = true
        }
        else if (assignment(x) == 0) {
          assume(lit, Propagation)

          for (i <- watchingClauses(x.id)) {
            assert(i >= 0)
            remaining ++= propagateClause(i)
          }
        }
      }
    }

    def allAssigned(): Boolean = Bench.time("allAssigned") {
      assignments.tail.forall(_ != 0)
    }

    def nextVar(lit: Lit): Option[VarId] = {
      val id = lit.getVar.id
      if (id < numVars) Some(VarId(id + 1))
      else None
    }

    def decide(lit: Lit): Unit = {
      if (unassigned(lit)) {
        assume(lit, Decision)
      }
      nextDecision = nextVar(lit).map(_.toLiteral(true))
    }

    // Attempt to backtrack and set the next decision to make
    // We pop the `trail` stack until we find a variable which was set to `true`
    // by a decision. When we find it, we set the next decision to be this
    // variable set to `false`.
    def backtrack(): Unit = Bench.time("backtrack") { trail match {
      case Nil =>
        // no more decisions to make, we have a conflict and we are unsat
        nextDecision = None
      case (lit, decision) :: lits =>
        trail = lits
        unassign(lit)
        if (decision == Decision && lit.isPositive) {
          conflict = false
          nextDecision = Some(-lit)
        } else
          backtrack()
    } }

    override def toString: String = {
      "\n\n============================\n" +
      "Assignment:\n" +
        assignments.zipWithIndex.tail.map { case (v,i) => "  " + i + " -> " + v }.mkString("\n") + "\n" +
      "Initial clauses:\n" +
        clauses.zipWithIndex.map { case (s, i) =>
          "  " + i + " -> " + s
        }.mkString("\n") + "\n" +
      "Watched Literals:\n" +
        watchedLiterals.zipWithIndex.map { case (s, i) =>
          "  " + i + " -> " + s.toList.mkString(" ")
        }.mkString("\n") +
      "\nNext decision to make: " + nextDecision +
      "\nConflict: " + conflict +
      "\nTrail: " + trail.mkString(" ") +
      "\n============================\n"
    }

  }
}
