import TransitionSystem._
import Expr._
import Utils._
import ScalaZ3Interface._

import sys.process._
import sys.error

import scala.language.postfixOps

case class TransitionSystem(
  val stateIds: Seq[StateIdentifier],
  val auxiliaryIds: Seq[AuxiliaryIdentifier],
  val inputIds: Seq[InputIdentifier],
  val transition: Map[Identifier, Expr],
) {

  checkWF()

  override def toString: String = {
    (auxiliaryIds ++ stateIds).map(id =>
      s"$id = ${transition(id)}"
    ).mkString("\n")
  }

  // Check some well-formedness conditions on the transition system
  def checkWF(): Unit = {
    // No two variables should have the same name
    val allVars = (stateIds ++ auxiliaryIds ++ inputIds).toArray
    for (i <- 0 to allVars.length - 1)
      for (j <- i + 1 to allVars.length - 1)
        if (allVars(i) == allVars(j))
         throw new Exception(s"Variable ${allVars(i)} is used twice in the transition system")

    // Every state variable must have an entry in the transition map
    stateIds.find(v => !transition.contains(v)) match {
      case Some(v) =>
        throw new Exception(s"State variable $v must have an entry in the transition map")
      case None => ()
    }

    // Every auxiliary variable must have an entry in the transition map
    auxiliaryIds.find(v => !transition.contains(v)) match {
      case Some(v) =>
        throw new Exception(s"Auxiliary variable $v must have an entry in the transition map")
      case None => ()
    }

    // Every auxiliary variable computation can only depend on previous auxiliary
    // variables (and the state and input variables)
    val auxiliaryArray = auxiliaryIds.toArray
    for (i <- 0 to auxiliaryArray.length - 1) {
      transition(auxiliaryArray(i)).fv.find(id =>
        !stateIds.contains(id) &&
        !inputIds.contains(id) &&
        !auxiliaryIds.take(i).contains(id)
      ) match {
        case None => ()
        case Some(id) =>
          throw new Exception(s"The computation of ${auxiliaryArray(i)} is not allowed to depend on $id")
      }
    }
  }

  def next(q: State, a: Input): State = {
    val env = evalVars(q ++ a, auxiliaryIds, transition)
    evalVarsWithAcc(env, Map(), stateIds, transition).collect {
      case (k: StateIdentifier, v) => (k, v)
    }
  }

  // Check whether there exist:
  // * a state `s0` that makes `from` evaluate to `true`,
  // * a state `sk` that makes `error` evaluate to `true`,
  // * a sequence of intermediary states `s1,...,s_{k-1}`
  // * a sequence of inputs `a0,...,a_{k-1}` such that for all i, next(s_i, a_i) = s_{i+1}
  //
  // When this is the case, the function returns
  // Some(Seq(s0, ..., sk), Seq(a0, ..., a_{k-1})), and
  // None otherwise.
  def checkErrorReachability(from: Expr, error: Expr, k: Int): Option[(Seq[State], Seq[Input])] = {
    type Fresh[A] = A

    // We build `k + 1` copies of `stateIds` to represent `s0, ..., sk`
    val newStateIds: Array[Seq[(StateIdentifier, Fresh[StateIdentifier])]] = (0 to k).toArray.map { i =>
      stateIds.map(s => (s, StateIdentifier(s"fresh_s_${i}_$s")))
    }
    // We build `k` copies of `inputIds` to represent `a0, ..., a_{k-1}`
    val newInputIds: Array[Seq[(InputIdentifier, Fresh[InputIdentifier])]] = (0 to k-1).toArray.map { i =>
      inputIds.map(a => (a, InputIdentifier(s"fresh_a_${i}_$a")))
    }
    // We build `k` copies of `auxiliaryIds` to represent computation variables
    // required to get `next(s_j, a_j)`
    val newAuxiliaryVars: Array[Seq[(AuxiliaryIdentifier, Fresh[AuxiliaryIdentifier])]] = (0 to k-1).toArray.map { i =>
      auxiliaryIds.map(c => (c, AuxiliaryIdentifier(s"fresh_c_${i}_$c")))
    }

    // We then build an expression `e` with three conjuncts stating that:
    // * for all j, next(s_j, a_j) = s_{j+1} (`eNext`)
    // * `s0` makes `from` evaluate to `true` (`eFrom`)
    // * `sk` makes `error` evaluate to `true` (`eError`)
    val eNext1: Expr = and(newAuxiliaryVars.zipWithIndex.flatMap {
      case (ids, i) =>
        ids.map {
          case (id, fid) => Equal(Var(fid), transition(id).subst(newStateIds(i).toMap ++ newInputIds(i).toMap ++ newAuxiliaryVars(i).toMap))
        }
    })
    val eNext2: Expr = and(newStateIds.zipWithIndex.tail.flatMap {
      case (ids, i) =>
        ids.map {
          case (id, fid) => Equal(Var(fid), transition(id).subst(newStateIds(i-1).toMap ++ newInputIds(i-1).toMap ++ newAuxiliaryVars(i-1).toMap))
        }
    })
    val eNext: Expr = And(eNext1, eNext2)
    val eFrom: Expr = from.subst(newStateIds(0).toMap)
    val eError: Expr = error.subst(newStateIds(k).toMap)
    val e = and(eNext, eFrom, eError)

    // We check whether `e` is satisfiable (and if so, we get a model):
    z3CheckSat(e) match {
      case None => None
      case Some(model) =>
        val sSeq = newStateIds.map(ids => ids.map { case (id, fid) => id -> model.getOrElse(fid, false) }.toMap )
        val aSeq = newInputIds.map(ids => ids.map { case (id, fid) => id -> model.getOrElse(fid, false) }.toMap )
        Some(sSeq, aSeq)
    }
  }

  def studentCheckErrorReachability(from: Expr, error: Expr, k: Int): Option[(Seq[State], Seq[Input])] = {
    ???
  }

  def checkErrorReachabilityAtMost(from: Expr, error: Expr, k: Int): Option[(Seq[State], Seq[Input])] = {
    def loop(i: Int): Option[(Seq[State], Seq[Input])] = {
      if (i <= k) checkErrorReachability(from, error, i) orElse loop(i + 1)
      else None
    }
    loop(0)
  }

  def checkReachability(from: State, errorState: State, k: Int): Option[(Seq[State], Seq[Input])] = {
    checkErrorReachability(stateToExpr(from), stateToExpr(errorState), k)
  }

  def checkReachabilityAtMost(from: State, errorState: State, k: Int): Option[(Seq[State], Seq[Input])] = {
    checkErrorReachabilityAtMost(stateToExpr(from), stateToExpr(errorState), k)
  }

  // Checks whether:
  // * `inv` is true in the `initial` states
  // * `inv` implies the `target` expression
  // * `inv` is inductive
  def checkInductiveInvariant(inv: Expr, initial: Expr, target: Expr): Boolean = {
    (checkErrorReachability(initial, Not(inv), 0) match {
      case None =>
        // println(" => OK: The invariant is initially valid")
        true
      case Some((sSeq, aSeq)) =>
        println("ERROR: The invariant is not initially valid")
        false
    }) &&
    (checkErrorReachability(inv, Not(target), 0) match {
      case None =>
        // println(" => OK: The inductive invariant implies the target expression")
        true
      case Some((sSeq, aSeq)) =>
        println("ERROR: The inductive invariant does not imply the target expression")
        false
    }) &&
    (checkErrorReachability(inv, Not(inv), 1) match {
      case None =>
        // println(" => OK: The invariant is inductive")
        true

      case Some((sSeq, aSeq)) =>
        // println("States")
        // println(sSeq.mkString("\n"))
        // println("Inputs")
        // println(aSeq.mkString("\n"))
        println("ERROR: The invariant is not inductive")
        false
    })
  }

  def dump(fileName: String, hlPath: Option[Path] = None, toPNG: Boolean = true) = {

    val baseName = fileName.replaceAll("\\.[^.]*$", "")
    val pngName = baseName + ".png"
    val dotName = baseName + ".dot"
    val out = new java.io.PrintStream(dotName)

    val allStates: Set[State] = enumerate(stateIds)
    val allInputs: Set[Input] = enumerate(inputIds)
    val intOfState: Map[State, Int] = allStates.zipWithIndex.toMap

    out.println("digraph dfa {")
    for ((q, i) <- intOfState) {
      if (hlPath.exists(path => path._2 == q || path._1.toSet.exists(_._1 == q))) {
        out.println(s"""\t$i [label = "${mapToString(q)}", fillcolor = green3, style = filled];""")
      } else {
        out.println(s"""\t$i [label = "${mapToString(q)}"];""")
      }
    }
    out.println()

    for ((p, i) <- intOfState) {
      for (input <- allInputs) {
        val q = next(p, input)
        if (hlPath.exists(path => path._1.toSet.contains((p, input)))) {
          out.println(s"""\t$i -> ${intOfState(q)} [ label = "${mapToString(input)}", color = crimson, fontcolor = crimson ];""")
        } else {
          out.println(s"""\t$i -> ${intOfState(q)} [ label = "${mapToString(input)}" ];""")
        }
      }
    }

    out.println("}")
    out.close
    if (toPNG)
      "dot -Tpng -o %s %s".format(pngName,dotName) !!;
  }

}

object TransitionSystem {
  type State = Map[StateIdentifier, Boolean]
  type Input = Map[InputIdentifier, Boolean]
  type Path = (Seq[(State, Input)], State)

  // After evaluating the expression of a given variable, this function saves
  // the result in an accumulator which is not visible when interpreting later
  // variables
  def evalVarsWithAcc(
    env: Map[Identifier, Boolean],
    acc: Map[Identifier, Boolean],
    vars: Seq[Identifier],
    transition: Map[Identifier, Expr]
  ): Map[Identifier, Boolean] = vars match {
    case Seq() => acc
    case x +: xs => evalVarsWithAcc(env, acc.updated(x, transition(x).interpret(env)), xs, transition)
  }

  // After evaluating the expression of a given variable, this function saves
  // the result in the environment so that it can be used for later variables
  def evalVars(
    env: Map[Identifier, Boolean],
    vars: Seq[Identifier],
    transition: Map[Identifier, Expr]
  ): Map[Identifier, Boolean] = vars match {
    case Seq() => env
    case x +: xs => evalVars(env.updated(x, transition(x).interpret(env)), xs, transition)
  }

  // Returns an expression whose only satisfying assignment is `s`
  def stateToExpr(s: State): Expr = {
    and(s.map { case (id, b) =>
      if (b) Var(id)
      else Not(Var(id))
    })
  }

  // Returns a random transition system
  def random(stateIds: Seq[StateIdentifier], auxiliaryIds: Seq[AuxiliaryIdentifier], inputIds: Seq[InputIdentifier]): TransitionSystem = {
    val auxiliaryTransition: Map[Identifier, Expr] = auxiliaryIds.foldLeft((stateIds ++ inputIds, Map[Identifier, Expr]())){
      case ((fv, acc), id) =>
        (fv :+ id, acc.updated(id, Expr.random(fv.toArray)))
    }._2
    val varArray: Array[Identifier] = (stateIds ++ inputIds ++ auxiliaryIds).toArray
    val stateTransition: Map[Identifier, Expr] = stateIds.map(id => id -> Expr.random(varArray)).toMap
    TransitionSystem(stateIds, auxiliaryIds, inputIds, auxiliaryTransition ++ stateTransition)
  }

}
