import Expr._

object Tseytin {

  def tseytin(e: Expr): Expr = {

    val fvs = e.fv
    for (fv <- fvs) {
      assert(
        !fv.id.startsWith("$$fresh"),
        s"You should not use a variable name starting with `$$fresh`: $fv"
      )
    }

    var fresh = fvs.size + 1 // next index which is not used
    def freshVar(): Var = {
      val res = Var(StateIdentifier("$$fresh" + fresh))
      fresh += 1
      res
    }

    val topLevelK: Expr = freshVar()

    // `k` is the new variable that we introduce for expression `e`
    def loop(e: Expr, k: Expr): Expr = {
      e match {
        case Var(id) =>
          And(
            Or(-k, e), // k => e
            Or(-e, k), // e => k
          )

        case BooleanLiteral(true) =>
          and(
            or(k), // k must be true
          )

        case BooleanLiteral(false) =>
          and(
            or(-k), // fresh must be false
          )

        case And(e1, e2) =>
          val k1 = freshVar()   // for e1
          val k2 = freshVar()   // for e2
          fresh += 2
          val cnf1 = loop(e1, k1)
          val cnf2 = loop(e2, k2)

          and(
            cnf1,
            cnf2,
            or(-k, k1),      // k ==> k1
            or(-k, k2),      // k ==> k2
            or(-k1, -k2, k), // k1 /\ k2 ==> k
          )

        case Or(e1, e2) =>
          val k1 = freshVar()   // for e1
          val k2 = freshVar()   // for e2
          fresh += 2
          val cnf1 = loop(e1, k1)
          val cnf2 = loop(e2, k2)

          and(
            cnf1,
            cnf2,
            or(-k1, k),      // k1 ==> k
            or(-k2, k),      // k2 ==> k
            or(-k, k1, k2),  // k ==> k1 \/ k2
          )

        case Xor(e1, e2) =>
          val k1 = freshVar()   // for e1
          val k2 = freshVar()   // for e2
          fresh += 2
          val cnf1 = loop(e1, k1)
          val cnf2 = loop(e2, k2)

          and(
            cnf1,
            cnf2,
            or(-k1, k2, k),   // k1 /\ not k2 ==> k
            or(k1, -k2, k),   // not k1 /\ k2 ==> k
            or(-k, k1, k2),   // k ==> k1 \/ k2
            or(-k, -k1, -k2), // k ==> not k1 \/ not k2
          )

        case Implies(e1, e2) =>
          val k1 = freshVar()   // for e1
          val k2 = freshVar()   // for e2
          fresh += 2
          val cnf1 = loop(e1, k1)
          val cnf2 = loop(e2, k2)

          and(
            cnf1,
            cnf2,
            or(-k1, k2, -k), // k1 /\ not k2 ==> not k
            or(k1, k),       // not k1 ==> k
            or(-k2, k),      // k2 ==> k
          )

        case Equal(e1, e2) =>
          val k1 = freshVar()   // for e1
          val k2 = freshVar()   // for e2
          fresh += 2
          val cnf1 = loop(e1, k1)
          val cnf2 = loop(e2, k2)

          and(
            cnf1,
            cnf2,
            or(-k1, k2, -k), // k1 /\ not k2 ==> not k
            or(k1, -k2, -k), // not k1 /\ k2 ==> not k
            or(-k1, -k2, k), // k1 /\ k2 ==> k
            or(k1, k2, k),   // not k1 /\ not k2 ==> k
          )

        case Not(e0) =>
          val k0 = freshVar()   // for e0
          fresh += 1
          val cnf0 = loop(e0, k0)

          and(
            cnf0,
            or(-k0, -k), // k0 ==> not k
            or(k0, k),   // not k0 ==> k
          )
      }
    }

    And(topLevelK, loop(e, topLevelK))
  }
}
